@@ -94,17 +94,22 @@ struct request_input_reply {
 static void request_input_passphrase_reply(DBusPendingCall *call, void *user_data)
 {
 	struct request_input_reply *passphrase_reply = user_data;
+	connman_bool_t values_received = FALSE;
 	connman_bool_t wps = FALSE;
 	char *identity = NULL;
 	char *passphrase = NULL;
 	char *wpspin = NULL;
 	char *key;
+	char *name = NULL;
+	int name_len = 0;
 	DBusMessageIter iter, dict;
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 
 	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
 		goto done;
 
+	values_received = TRUE;
+
 	dbus_message_iter_init(reply, &iter);
 	dbus_message_iter_recurse(&iter, &dict);
 	while (dbus_message_iter_get_arg_type(&dict) == DBUS_TYPE_DICT_ENTRY) {
@@ -141,6 +146,28 @@ static void request_input_passphrase_reply(DBusPendingCall *call, void *user_dat
 			dbus_message_iter_recurse(&entry, &value);
 			dbus_message_iter_get_basic(&value, &wpspin);
 			break;
+		} else if (g_str_equal(key, "Name")) {
+			dbus_message_iter_next(&entry);
+			if (dbus_message_iter_get_arg_type(&entry)
+							!= DBUS_TYPE_VARIANT)
+				break;
+			dbus_message_iter_recurse(&entry, &value);
+			dbus_message_iter_get_basic(&value, &name);
+			name_len = strlen(name);
+		} else if (g_str_equal(key, "SSID")) {
+			dbus_message_iter_next(&entry);
+			if (dbus_message_iter_get_arg_type(&entry)
+							!= DBUS_TYPE_VARIANT)
+				break;
+			dbus_message_iter_recurse(&entry, &value);
+			if (dbus_message_iter_get_arg_type(&value)
+							!= DBUS_TYPE_VARIANT)
+				break;
+			if (dbus_message_iter_get_element_type(&value)
+							!= DBUS_TYPE_VARIANT)
+				break;
+			dbus_message_iter_get_fixed_array(&value, &name,
+							&name_len);
 		}
 		dbus_message_iter_next(&dict);
 	}
@@ -164,8 +191,10 @@ static void request_input_passphrase_reply(DBusPendingCall *call, void *user_dat
 	}
 
 done:
-	passphrase_reply->callback(passphrase_reply->service, identity,
-				passphrase, passphrase_reply->user_data);
+	passphrase_reply->callback(passphrase_reply->service, values_received,
+				name, name_len,
+				identity, passphrase,
+				passphrase_reply->user_data);
 	connman_service_unref(passphrase_reply->service);
 	dbus_message_unref(reply);
 	g_free(passphrase_reply);
@@ -198,7 +227,7 @@ static void request_input_append_identity(DBusMessageIter *iter,
 
 	connman_dbus_dict_append_basic(iter, "Type",
 				DBUS_TYPE_STRING, &str);
-	str = "Mandatory";
+	str = "mandatory";
 	connman_dbus_dict_append_basic(iter, "Requirement",
 				DBUS_TYPE_STRING, &str);
 }
@@ -234,7 +263,7 @@ static void request_input_append_passphrase(DBusMessageIter *iter,
 	}
 	connman_dbus_dict_append_basic(iter, "Type",
 				DBUS_TYPE_STRING, &value);
-	value = "Mandatory";
+	value = "mandatory";
 	connman_dbus_dict_append_basic(iter, "Requirement",
 				DBUS_TYPE_STRING, &value);
 
@@ -257,6 +286,32 @@ static void request_input_append_wps(DBusMessageIter *iter, void *user_data)
 				DBUS_TYPE_STRING, &str);
 }
 
+static void request_input_append_name(DBusMessageIter *iter, void *user_data)
+{
+	const char *str = "string";
+
+	connman_dbus_dict_append_basic(iter, "Type",
+				DBUS_TYPE_STRING, &str);
+	str = "mandatory";
+	connman_dbus_dict_append_basic(iter, "Requirement",
+				DBUS_TYPE_STRING, &str);
+	connman_dbus_dict_append_array(iter, "Alternates",
+				DBUS_TYPE_STRING,
+				request_input_append_alternates,
+				"SSID");
+}
+
+static void request_input_append_ssid(DBusMessageIter *iter, void *user_data)
+{
+	const char *str = "ssid";
+
+	connman_dbus_dict_append_basic(iter, "Type",
+				DBUS_TYPE_STRING, &str);
+	str = "alternate";
+	connman_dbus_dict_append_basic(iter, "Requirement",
+				DBUS_TYPE_STRING, &str);
+}
+
 static void request_input_append_password(DBusMessageIter *iter,
 							void *user_data)
 {
@@ -264,7 +319,7 @@ static void request_input_append_password(DBusMessageIter *iter,
 
 	connman_dbus_dict_append_basic(iter, "Type",
 				DBUS_TYPE_STRING, &str);
-	str = "Mandatory";
+	str = "mandatory";
 	connman_dbus_dict_append_basic(iter, "Requirement",
 				DBUS_TYPE_STRING, &str);
 }
@@ -314,6 +369,7 @@ static void request_input_login_reply(DBusPendingCall *call, void *user_data)
 
 done:
 	username_password_reply->callback(username_password_reply->service,
+					TRUE, NULL, 0,
 					username, password,
 					username_password_reply->user_data);
 	connman_service_unref(username_password_reply->service);
@@ -348,6 +404,13 @@ int __connman_agent_request_passphrase_input(struct connman_service *service,
 
 	connman_dbus_dict_open(&iter, &dict);
 
+	if (__connman_service_is_hidden(service)) {
+		connman_dbus_dict_append_dict(&dict, "Name",
+					request_input_append_name, NULL);
+		connman_dbus_dict_append_dict(&dict, "SSID",
+					request_input_append_ssid, NULL);
+	}
+
 	if (__connman_service_get_security(service) ==
 			CONNMAN_SERVICE_SECURITY_8021X) {
 		connman_dbus_dict_append_dict(&dict, "Identity",
@@ -461,6 +524,91 @@ int __connman_agent_request_login_input(struct connman_service *service,
 	return -EIO;
 }
 
+struct request_browser_reply_data {
+	struct connman_service *service;
+	browser_authentication_cb_t callback;
+	void *user_data;
+};
+
+static void request_browser_reply(DBusPendingCall *call, void *user_data)
+{
+	struct request_browser_reply_data *browser_reply_data = user_data;
+	DBusMessage *reply = dbus_pending_call_steal_reply(call);
+	connman_bool_t result = FALSE;
+
+	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
+		goto done;
+
+	result = TRUE;
+
+done:
+	browser_reply_data->callback(browser_reply_data->service,
+					result,	browser_reply_data->user_data);
+	connman_service_unref(browser_reply_data->service);
+	dbus_message_unref(reply);
+	g_free(browser_reply_data);
+}
+
+int __connman_agent_request_browser(struct connman_service *service,
+				browser_authentication_cb_t callback,
+				const char *url, void *user_data)
+{
+	struct request_browser_reply_data *browser_reply_data;
+	DBusPendingCall *call;
+	DBusMessage *message;
+	DBusMessageIter iter;
+	const char *path;
+
+	if (service == NULL || agent_path == NULL || callback == NULL)
+		return -ESRCH;
+
+	if (url == NULL)
+		url = "";
+
+	message = dbus_message_new_method_call(agent_sender, agent_path,
+					CONNMAN_AGENT_INTERFACE,
+					"RequestBrowser");
+	if (message == NULL)
+		return -ENOMEM;
+
+	dbus_message_iter_init_append(message, &iter);
+
+	path = __connman_service_get_path(service);
+	dbus_message_iter_append_basic(&iter, DBUS_TYPE_OBJECT_PATH, &path);
+
+	dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &url);
+
+	browser_reply_data = g_try_new0(struct request_browser_reply_data, 1);
+	if (browser_reply_data == NULL) {
+		dbus_message_unref(message);
+		return -ENOMEM;
+	}
+
+	if (dbus_connection_send_with_reply(connection, message,
+							&call, -1) == FALSE) {
+		dbus_message_unref(message);
+		g_free(browser_reply_data);
+		return -ESRCH;
+	}
+
+	if (call == NULL) {
+		dbus_message_unref(message);
+		g_free(browser_reply_data);
+		return -ESRCH;
+	}
+
+	browser_reply_data->service = connman_service_ref(service);
+	browser_reply_data->callback = callback;
+	browser_reply_data->user_data = user_data;
+
+	dbus_pending_call_set_notify(call, request_browser_reply,
+						browser_reply_data, NULL);
+
+	dbus_message_unref(message);
+
+	return -EIO;
+}
+
 struct report_error_data {
 	struct connman_service *service;
 	report_error_cb_t callback;
