@@ -227,9 +227,13 @@ static struct gateway_data *add_gateway(struct connman_service *service,
 			data->ipv4_gateway = old->ipv4_gateway;
 			old->ipv4_gateway = NULL;
 		}
+	} else {
+		/*
+		 * Only take a ref if we are adding new stuff to hash.
+		 */
+		connman_service_ref(service);
 	}
 
-	connman_service_ref(service);
 	g_hash_table_replace(gateway_hash, service, data);
 
 	return data;
