@@ -64,6 +64,8 @@ struct connman_wispr_portal_context {
 
 	const char *status_url;
 
+	char *redirect_url;
+
 	/* WISPr specific */
 	GWebParser *wispr_parser;
 	struct connman_wispr_message wispr_msg;
@@ -130,6 +132,8 @@ static void free_connman_wispr_portal_context(struct connman_wispr_portal_contex
 
 	g_web_unref(wp_context->web);
 
+	g_free(wp_context->redirect_url);
+
 	g_web_parser_unref(wp_context->wispr_parser);
 	connman_wispr_message_init(&wp_context->wispr_msg);
 
@@ -429,6 +433,8 @@ static gboolean wispr_input(const guint8 **data, gsize *length,
 }
 
 static void wispr_portal_request_wispr_login(struct connman_service *service,
+				connman_bool_t success,
+				const char *ssid, int ssid_len,
 				const char *username, const char *password,
 				void *user_data)
 {
@@ -484,8 +490,6 @@ static gboolean wispr_manage_message(GWebResult *result,
 
 		wp_context->wispr_result = CONNMAN_WISPR_RESULT_LOGIN;
 
-		__connman_service_request_login(wp_context->service);
-
 		if (__connman_agent_request_login_input(wp_context->service,
 					wispr_portal_request_wispr_login,
 					wp_context) != -EIO)
@@ -520,6 +524,24 @@ static gboolean wispr_manage_message(GWebResult *result,
 	return FALSE;
 }
 
+static void wispr_portal_browser_reply_cb(struct connman_service *service,
+					connman_bool_t authentication_done,
+					void *user_data)
+{
+	struct connman_wispr_portal_context *wp_context = user_data;
+
+	if (service == NULL || wp_context == NULL)
+		return;
+
+	if (authentication_done == FALSE) {
+		wispr_portal_error(wp_context);
+		return;
+	}
+
+	/* Restarting the test */
+	__connman_wispr_start(service, wp_context->type);
+}
+
 static gboolean wispr_portal_web_result(GWebResult *result, gpointer user_data)
 {
 	struct connman_wispr_portal_context *wp_context = user_data;
@@ -564,22 +586,32 @@ static gboolean wispr_portal_web_result(GWebResult *result, gpointer user_data)
 								&str) == TRUE)
 			portal_manage_status(result, wp_context);
 		else
-			__connman_service_request_login(wp_context->service);
+			__connman_agent_request_browser(wp_context->service,
+					wispr_portal_browser_reply_cb,
+					wp_context->redirect_url, wp_context);
 
 		break;
 	case 302:
 		if (g_web_result_get_header(result, "Location",
-						&redirect) == FALSE)
+						&redirect) == FALSE) {
+			__connman_agent_request_browser(wp_context->service,
+					wispr_portal_browser_reply_cb,
+					wp_context->status_url, wp_context);
 			break;
+		}
 
 		DBG("Redirect URL: %s", redirect);
 
+		wp_context->redirect_url = g_strdup(redirect);
+
 		wp_context->request_id = g_web_request_get(wp_context->web,
 				redirect, wispr_portal_web_result, wp_context);
 
 		goto done;
 	case 404:
-		wispr_portal_error(wp_context);
+		if (__connman_service_online_check_failed(wp_context->service,
+							wp_context->type) == 0)
+			wispr_portal_error(wp_context);
 
 		break;
 	default:
@@ -627,8 +659,10 @@ static int wispr_portal_detect(struct connman_wispr_portal_context *wp_context)
 {
 	enum connman_service_type service_type;
 	char *interface = NULL;
+	char **nameservers = NULL;
 	int if_index;
 	int err = 0;
+	int i;
 
 	DBG("wispr/portal context %p", wp_context);
 	DBG("service %p", wp_context->service);
@@ -657,8 +691,16 @@ static int wispr_portal_detect(struct connman_wispr_portal_context *wp_context)
 	DBG("interface %s", interface);
 
 	if_index = connman_inet_ifindex(interface);
-	if (if_index < 0)
-		return -EINVAL;
+	if (if_index < 0) {
+		err = -EINVAL;
+		goto done;
+	}
+
+	nameservers = connman_service_get_nameservers(wp_context->service);
+	if (nameservers == NULL) {
+		err = -EINVAL;
+		goto done;
+	}
 
 	wp_context->web = g_web_new(if_index);
 	if (wp_context->web == NULL) {
@@ -674,6 +716,9 @@ static int wispr_portal_detect(struct connman_wispr_portal_context *wp_context)
 		wp_context->status_url = STATUS_URL_IPV6;
 	}
 
+	for (i = 0; nameservers[i] != NULL; i++)
+		g_web_add_nameserver(wp_context->web, nameservers[i]);
+
 	wp_context->token = connman_proxy_lookup(interface,
 					wp_context->status_url,
 					wp_context->service,
@@ -682,6 +727,8 @@ static int wispr_portal_detect(struct connman_wispr_portal_context *wp_context)
 		err = -EINVAL;
 
 done:
+	g_strfreev(nameservers);
+
 	g_free(interface);
 	return err;
 }
